normByLevels<-function(col1,col2)
{
  sums<-tapply(col1,factor(col2),sum) 
  sums[sums==0]<-1
  levs<-tapply(col1,factor(col2))
  col1/sums[][levs]
}

replaceNA<-function(col1,col2)
{
   col1[is.na(col1)] <- col2[which(is.na(col1))]
   col1
}

findhot<-function (dataFiles,range= seq(1,100,1),nullOffset=TRUE, 
    deg= TRUE,    rand= TRUE,
   silent=TRUE,pfp= FALSE, single=FALSE, double= FALSE,delta= 0.048, hotfall= 0,tri= FALSE)
{
  for (i in range) {
    g<-linearFETA(dataFiles,nullOffset=nullOffset,deg=deg,rand=rand, 
        silent=silent,pfp=pfp, single=single, 
        double=double,delta=delta, hotcount=i, hotfall=hotfall, tri=tri)
    cat (i," ",g$deviance)

    cat ("\n")
    rm(g)
  }
}

findhotfall<-function (dataFiles,range= seq(1,100,1),nullOffset=TRUE, 
    deg= TRUE,    rand= TRUE, hotcount=0,
   silent=TRUE,pfp= FALSE, single=FALSE, double= FALSE,delta= 0.048, tri= FALSE)
{
  for (i in range) {
    g<-linearFETA(dataFiles,nullOffset=nullOffset,deg=deg,rand=rand, 
        silent=silent,pfp=pfp, single=single, 
        double=double,delta=delta, hotcount=hotcount, tri=tri,hotfall=i)
    cat (i," ",g$deviance)

    cat ("\n")
    rm(g)
  }
}

finddelta<-function (dataFiles,range= seq(0.01,0.1,0.01),nullOffset=TRUE, 
    deg= TRUE,    rand= TRUE,
   silent=TRUE,pfp=TRUE, single=FALSE, double= FALSE,delta= 0.048, tri= FALSE,  
   hotcount= 0, hotfall=0)
{
  for (i in range) {
    g<-linearFETA(dataFiles,nullOffset=nullOffset,deg=deg,rand=rand, 
        silent=silent,pfp=pfp, single=single, 
        double=double,delta=i, tri=tri,hotcount=hotcount,hotfall=hotfall)
    cat (i," ",g$deviance)
   
    
    cat ("\n")
    rm(g)
  }
}


finddeltagamma<-function (dataFiles,deltarange= seq(0.01,0.1,0.01),gammarange=seq(0.01,0.1,0.01),nullOffset=TRUE,
    deg= FALSE, rand= FALSE,
   silent=TRUE,pfp= FALSE, pfp2=TRUE, single=FALSE, double= FALSE,delta= 0.048, hotcount=0, gamma=0, tri= FALSE,
   family= FALSE)
{     
  for (i in deltarange) {
    for (j in gammarange) {
    	g<-linearFETA(dataFiles,nullOffset=nullOffset,deg=deg,rand=rand,
        	silent=silent,pfp=pfp, pfp2=pfp2, single=single,
        	double=double,delta=i, gamma=j, hotcount= hotcount, tri=tri,family=family)
    	cat ("delta=", i,", gamma=",j, g$deviance,"\n")
    	rm(g)
    }
  }
} 


find_all_deltas<-function (dataFiles,delta_1_range= seq(0.01,0.1,0.01), delta_2_range=seq(0.01,0.1,0.01),
   delta_3_range = seq(0.01,0.1,0.01), gammarange = seq(0.01,0.1,0.01), nullOffset=TRUE,
   deg= FALSE, rand= FALSE,
   silent=TRUE,pfp= FALSE, pfp2=FALSE, pfp3 = TRUE, single=FALSE, double= FALSE,delta= 0.048, gamma = 0.0, tri= FALSE, 
   delta_1 = 1.0, delta_2 = 0.048, delta_3 = 0.0, 
   family= FALSE)
{
  for (i in delta_1_range) {
    for (j in delta_2_range) {
       for (k in delta_3_range) {
          g<-linearFETA(dataFiles,nullOffset=nullOffset,deg=deg,rand=rand,
                silent=silent,pfp=pfp, pfp2=pfp2, pfp3=pfp3, single=single,
                double=double,delta_1=i, delta_2=j, delta_3=k, tri=tri,
                family=family)
          cat ("delta_1=",i,",delta_2=",j,",delta_3=",k,",",g$deviance,"\n")
          rm(g)
       }
    }
  }
}


FETAread<-function(dataFiles)
{
   flag<-0
   
   for (d in dataFiles) {
      if (flag == 0) {
         flag <-1
         readdata<-read.table(d,col.names=c("nodeid","choiceid","timeschosen","batchsize",
           "nonodes", "degcount","tricount",
           "defPFP","degsum","singlecount","doublecount", "trisum", "pfpsum", "hotcount"))
      } else {
         readdata<-rbind(readdata,read.table(d))
      }
   }
   readdata
}

linearFETA<-function (dataFiles,nullOffset=TRUE, deg= TRUE, rand= TRUE,
   silent=FALSE,pfp= FALSE, pfp2=FALSE, pfp3=FALSE, antitri=FALSE,
   single=FALSE, double= FALSE,
   delta= 0.048, gamma = 0, tri= FALSE, 
   family= FALSE, delta_1= 1.0, delta_2=0.048, delta_3 = 0.0, hotcount= 0, hotfall=0,newcount=0)
{
  scount<-0
  readdata<-FETAread(dataFiles)
  if (!family) {
    if (max(readdata$batchsize) == 1) {
      family<-binomial(link="identity")
    } else {
      family<-gaussian(link="identity")
    }
  }
  if (!silent) {
    cat("Data read\n")
  }
  prob<-readdata$timeschosen
  if (nullOffset) {
    myForm<-"prob ~ 0"
  } else {
    scount<-1
    myForm<-"prob ~ 1"
  }
  if (rand) {
    myForm<-paste(myForm,"+ randFact")
    randFact<-readdata$batchsize/readdata$nonodes
    scount<-scount+1
  }
  if (hotcount > 0) {
     hotcol<-1*(readdata$hotcount < hotcount)
     hotcol<-normByLevels(hotcol, readdata$choiceid)
     myForm<-paste(myForm,"+ hotcol")
     scount<-scount+1
  } 
  if (newcount > 0) {
     newcol<-1*( (max(readdata$nodeid) -readdata$nodeid) < newcount)
     newcol<-normByLevels(newcol, readdata$choiceid)
     myForm<-paste(myForm,"+ newcol")
     scount<-scount+1
  }
  if (hotfall > 0) {
     hotfallcol<-(hotfall - readdata$hotcount)*(readdata$hotcount < hotfall)
     hotfallcol<-normByLevels(hotfallcol,readdata$choiceid)
     myForm<-paste(myForm,"+ hotfallcol")
     scount<-scount+1
  }
  if (deg) {
     degrees<-readdata$batchsize*readdata$degcount/readdata$degsum
     myForm<-paste(myForm,"+ degrees")
     scount<-scount+1
  }
#  sums<-tapply(degrees,factor(readdata$V3),sum)
#  cat(sums) 
  if (pfp) {
    pfpcol<-readdata$degcount^(1.0 + delta * log10(readdata$degcount))
    pfpcol<-normByLevels(pfpcol,readdata$choiceid)
    pfpcol<-pfpcol*readdata$batchsize
    myForm<-paste(myForm,"+ pfpcol")
    scount<-scount+1
  }
  if (pfp2) {
    pfp2col<-readdata$degcount^(1.0 + delta * log10(readdata$degcount) * (1.0 + gamma * log10(readdata$degcount)))
    pfp2col<-normByLevels(pfp2col,readdata$choiceid)
    pfp2col<-pfp2col*readdata$batchsize
    myForm<-paste(myForm,"+ pfp2col")                                                                     
    scount<-scount+1                                                                                     
  }   
  if (pfp3) {                                                                                            
    pfp3col<-readdata$degcount^(delta_1 + delta_2 * log10(readdata$degcount) + delta_3 * log10(readdata$degcount) * log10(readdata$degcount))     
    pfp3col<-normByLevels(pfp3col,readdata$choiceid)                                                     
    pfp3col<-pfp3col*readdata$batchsize                                                                  
    myForm<-paste(myForm,"+ pfp3col")                                                                     
    scount<-scount+1                                                                                     
  }
  if (tri) {
    tricol<-readdata$tricount/readdata$trisum
    tricol<-replaceNA(tricol, 1/readdata$nonodes)
    myForm<-paste(myForm,"+ tricol")
    scount<-scount+1
    tricol<-tricol*readdata$batchsize
  }
  if (antitri) {
    antitricol<-max(readdata$tricount)-(readdata$tricount)
    antitricol<-normByLevels(antitricol,readdata$choiceid)
    myForm<-paste(myForm,"+ antitricol")
    scount<-scount+1
  }
  if (single) {
    singlecol<-(1*(readdata$degcount==1))*1/readdata$singlecount
    singlecol<-replaceNA(singlecol, 1/readdata$nonodes)
    singlecol<-singlecol*readdata$batchsize
    myForm<-paste(myForm,"+ singlecol")
    scount<-scount+1
  }
  if (double) {
    doublecol<-(1*(readdata$degcount==2))/readdata$doublecount
    doublecol<-replaceNA(doublecol, 1/readdata$nonodes)
    doublecol<-doublecol*readdata$batchsize
    myForm<-paste(myForm,"+ doublecol")
    scount<-scount+1
  }  
#  cat(degrees)
#  cat("\n\n\n")
#  cat(pfpcol)
#  cat("\n\n\n")

  if(!silent) {
    cat ("Now fitting ",myForm," with ",scount," variables\n")
  }
  fmla<-as.formula(myForm)
  mystart<-rep(1/scount,scount)
  g<-glm(fmla,family=family,start=mystart)
  g
}


