#!/usr/bin/env python

# Reads files of network information writes files of link and node additions
import sys, getopt, math
from netmodel import NetModel, NetData
import random

modelStep= 1


def printUsage():
  print >> sys.stderr, "netanalyser.py [options] [network file] [model file] [node file] [edge file]"
  print >> sys.stderr, "\t\t Produces R ready files for input"
  print >> sys.stderr, "\t\t -h this help"
  print >> sys.stderr, "\t\t -s print likelihood stats"
  print >> sys.stderr, "\t\t -S print likelihood stats verbosely"
  print >> sys.stderr, "\t\t -w [no iter] -- number of warm up iterations"
  print >> sys.stderr, "\t\t -t [time] -- start time"
  print >> sys.stderr, "\t\t -r [probability] -- take random samples with prob"
  print >> sys.stderr, "\t\t -m [no iter] -- max iter to output"
  print >> sys.stderr, "\t\t -O -- estimate outer model"
  print >> sys.stderr, "\t\t -o [filename] -- write explicit outer model"
  print >> sys.stderr, "\t\t -b [batch] -- process in batches of a given size"
  print >> sys.stderr, "\t\t -B -- process in batches according to time stamp"
  print >> sys.stderr, "\t\t network file is a list of edges from a network in time order"
  print >> sys.stderr, "\t\t model file is a description of the creation model (see README)"
  print >> sys.stderr, "\t\t node file is an output file with data for new nodes"
  print >> sys.stderr, "\t\t edge file is an output file with data for edges between old nodes"
  sys.exit()

nargs= len(sys.argv)
if nargs < 5:
  printUsage()
  sys.exit()
try:
  optlist, args= getopt.getopt(sys.argv[1:],"hsSBOo:w:m:t:r:b:")
except getopt.GetoptError, e:
  print >> sys.stderr, 'Option parsing error: '+str(e)+'.'
if len(args) != 4:
  printUsage()
warmUp= 0
startTime= -1
maxIter= -1
randSkip= 1
batchSize= 1
lStats= False
vStats= False
batch= False
batchTime= False
outerEst= False
outerModFile= False
random.seed()
for op in optlist:
  if op[0] == '-h':
    printUsage()
  elif op[0] == '-s':
    lStats= True
  elif op[0] == '-S':
    lStats= True
    vStats= True
  elif op[0] == '-B':
    batch= True
    batchTime= True
  elif op[0] == '-O':
    outerEst= True
  elif op[0] == '-w':
    try:
      warmUp= int(op[1])
    except:
      print >> sys.stderr,"-w expects intger argument"
      sys.exit()
    if warmUp < 0:
      print >> sys.stderr,"-w expects non negative integer"
      sys.exit()
  elif op[0] == '-o':
    outerModFile= op[1]
  elif op[0] == '-b':
    try:
      batchSize= int(op[1])
    except:
      print >> sys.stderr,"-b expects integer argument"
      sys.exit()
    if batchSize <= 0:
      print >> sys.stderr,"-b expects integer argument"
      sys.exit()
    batch= True
  elif op[0] == '-m':
    try:
      maxIter= int(op[1])
    except:
      print >> sys.stderr,"-m expects integer argument"
      sys.exit()
    if maxIter < 1:
      print >> sys.stderr,"-m expects positive integer"
      sys.exit()
  elif op[0] == '-t':
    try:
      startTime= int(op[1])
    except: 
      print >> sys.stderr,"-t expects integer argument"
      sys.exit()
    if startTime < 0:
      print >> sys.stderr,"-t expects positive integer"
      sys.exit()
  elif op[0] == '-r':
    try:
      randSkip= float(op[1])
      assert (randSkip >= 0)
      assert (randSkip < 1)
    except:
      print >> sys.stderr, "-r expects a float in range [0,1)"
      sys.exit()
# Give a default warm up
if outerModFile: 
  try:
    outerModPtr= open(outerModFile,"w")
  except:
    print >> sys.stderr, "Cannot write to outerModFile", outerModFile
    sys.exit()
if warmUp == 0 and startTime == -1:
  warmUp= 5
netFileName= args[0]
modelFileName= args[1]
nodeFileName= args[2]
edgeFileName= args[3]
model= NetModel()
model.read(modelFileName)
network= NetData()
network.read(netFileName)
try:
  nodeFile= open(nodeFileName,'w')
  edgeFile= open(edgeFileName,'w')
except Exception, e:
  print >> sys.stderr, "File error",e
  sys.exit()
written= 0
i= 1
network2= NetData(model)
nodeEdgeCount= 0
innerNodeCount= 0
batchEdgeCount= 0
batchNodeCount= 0
nodeLikelihood= 0.0         #likelihood estimators
randNodeLikelihood= 0.0     
edgeLikelihood= 0.0
edgeLikeErr= False
nodeLikeErr= False
randEdgeLikelihood= 0.0
if batch == True:
  nodeConn= network.nodeCount*[0]
  edgeConn= network.nodeCount*[0]
  currTime= -1
for e in network.edges[1:]:
  if startTime > 0:
    if e[2] > startTime:
      warmUp-=1;
  else:
    warmUp-=1;
  if warmUp < 0:
    break
  i+= 1
  if (e[0] >= network2.nodeCount):
    network2.addNode(e[0])
  network2.addEdge(e[0],e[1],e[2],model.nodePFPparm, model.edgePFPparm)
edgeChoiceNo= 1
nodeChoiceNo= 1
eStatCount= 0
nStatCount= 0
randNo= 0
outerInnerEdgeCount= 0
outerNewNodeCount= 0
outerLongestRun= 100
outerFirstNode= True
outerNewNodeArr= (outerLongestRun+1)*[0]
outerInnerEdgeArr= (outerLongestRun+1)*[0]
for e in network.edges[i:]:
  if randSkip < 1:
    randNo= random.random() 
  if batchTime == True:
    if currTime == -1:
      currTime= e[2]
    elif currTime != e[2]:
      model.printNodeBatch(nodeConn,network2,nodeChoiceNo,nodeFile)
      nodeChoiceNo+= 1
      model.printEdgeBatch(edgeConn,network2,edgeChoiceNo,edgeFile)
      edgeChoiceNo+= 1
      nodeConn= network.nodeCount*[0]
      edgeConn= network.nodeCount*[0]
      currTime= e[2]
  if (e[0] >= network2.nodeCount):
    if outerModFile:
      print >> outerModPtr, model.NET_MODEL_ADD_NODE,e[2]
    network2.addNode(e[0])
#    print outerInnerEdgeCount
    outerInnerEdgeArr[min(outerInnerEdgeCount,outerLongestRun)]+= 1 
    outerInnerEdgeCount= 0
    if outerNewNodeCount != 0:
      outerNewNodeArr[min(outerNewNodeCount,outerLongestRun)]+= 1
      outerNewNodeCount= 0
  if e[0] == network2.nodeCount-1:   # node connects to new node
    if outerModFile:
      print >> outerModPtr, model.NET_MODEL_ADD_NEW_NODE_EDGE,e[2]
    outerNewNodeCount+= 1
    if lStats and not nodeLikeErr:
      nStatCount+=1 
      (prob,rprob)= model.getNodeProb(e[1], network2)
      if prob <= 0.0:
        nodeLikeErr= True
      else:
        nodeLikelihood+= math.log(prob)
        randNodeLikelihood+= math.log(rprob)
    if batch:
      nodeConn[e[1]]+= 1
      batchNodeCount+= 1
    elif randNo < randSkip:
      model.printNodeChoice(e[1], network2, nodeChoiceNo, nodeFile)
      nodeChoiceNo+= 1
  else:   # node is between internal edges
    if outerModFile:
      print >> outerModPtr, model.NET_MODEL_ADD_INNER_EDGE,e[2]
    if outerNewNodeCount != 0:
      outerNewNodeArr[min(outerNewNodeCount,outerLongestRun)]+= 1
      outerNewNodeCount= 0
    outerInnerEdgeCount+= 1
    if random.random() < 0.5:  #Randomly swap nodes in case there was an order
      e= (e[1],e[0],e[2])
    if lStats and not edgeLikeErr:
      eStatCount+= 2
      (p1,rp1)= model.getEdgeProb(e[0], e[1], network2)
      (p2,rp2)= model.getEdgeProb(e[1], e[0], network2)
      prob= p1+p2
      rprob= rp1+rp2
      if prob <= 0.0:
        edgeLikeErr= True
      else:
        edgeLikelihood+= math.log(prob)
        randEdgeLikelihood+= math.log(rprob)
    if batch:
      edgeConn[e[0]]+= 1
      edgeConn[e[1]]+= 1
      batchEdgeCount+=2
    else:
      skiplist= [network2.nodeCount - 1]
      if randNo < randSkip:
        model.printEdgeChoice(e[0], network2, edgeChoiceNo, edgeFile,skiplist)
      edgeChoiceNo+= 1
      skiplist= [e[0]]
      if model.simpleGraph:
        skiplist.extend(network2.outNodes[e[0]])
      if not (network2.nodeCount - 1) in skiplist:
        skiplist.append(network2.nodeCount - 1)
      if randNo < randSkip:
        model.printEdgeChoice(e[1], network2, edgeChoiceNo, edgeFile,skiplist)
        edgeChoiceNo+= 1
  if batch and batchSize > 1:
    if batchNodeCount >= batchSize:
      model.printNodeBatch(nodeConn,network2,nodeChoiceNo,nodeFile)
      nodeChoiceNo+= 1
      batchNodeCount= 0
      nodeConn= network.nodeCount*[0]  
    if batchEdgeCount >= batchSize:
      model.printEdgeBatch(edgeConn,network2,edgeChoiceNo,edgeFile)
      edgeChoiceNo+= 1
      batchEdgeCount= 0
      edgeConn= network.nodeCount*[0]  
  network2.addEdge(e[0],e[1],e[2],model.nodePFPparm, model.edgePFPparm)
if batch:
  model.printNodeBatch(nodeConn,network2,nodeChoiceNo,nodeFile)
  model.printEdgeBatch(edgeConn,network2,edgeChoiceNo,edgeFile)
if outerEst:
  if model.simpleGraph:
    print "S"
  while len(outerNewNodeArr) != 0 and  outerNewNodeArr[-1] == 0:
    outerNewNodeArr= outerNewNodeArr[0:-1]
  while len(outerInnerEdgeArr) != 0 and  outerInnerEdgeArr[-1] == 0:
    outerInnerEdgeArr= outerInnerEdgeArr[0:-1]
  if len(outerNewNodeArr) !=0:
    print "n",
    for s in outerNewNodeArr:
      print float(s)/sum(outerNewNodeArr),
    print
  if len(outerInnerEdgeArr) !=0:
    print "e",
    for s in outerInnerEdgeArr:
      print float(s)/sum(outerInnerEdgeArr),
    print    
if lStats:
  if nodeLikeErr:
    if vStats:
      print "No node stats -- negative or zero probabilities occurred"
    else:
      print "NA NA NA NA NA"
  else:
    if vStats: 
      print "NODE MODEL"
      print "Log likelihood ",
    print nodeLikelihood, 
    if vStats:
      print "\nNull likelihood ",
    print randNodeLikelihood,
    if vStats:
      print "\nDeviance",
    print -2.0*nodeLikelihood,
    if vStats:
      print "\nNull deviance",
    print -2.0*(nodeLikelihood - randNodeLikelihood),
    if vStats:
      print "\nPer choice likelihood ratio",
    print math.exp((nodeLikelihood - randNodeLikelihood)/nStatCount),
    if vStats:
      print
  if edgeLikeErr:
    if vStats:
      print "No edge stats -- negative or zero probabilities occurred"
    else:
      print "NA NA NA NA NA",
  elif eStatCount == 0:
    if vStats:
      print "No edge stats -- no inner edges"
    else:
      print "NA NA NA NA NA",
  else:
    if vStats: 
      print "EDGE MODEL"
      print "Log likelihood ",
    print edgeLikelihood, 
    if vStats:
      print "\nNull likelihood ",
    print randEdgeLikelihood,
    if vStats:
      print "\nDeviance",
    print -2.0*edgeLikelihood,
    if vStats:
      print "\nNull deviance",
    print -2.0*(edgeLikelihood - randEdgeLikelihood),
    if vStats:
      print "\nPer choice likelihood ratio",
    print math.exp((edgeLikelihood - randEdgeLikelihood)/eStatCount),
    if vStats:
      print
  if not nodeLikeErr and not edgeLikeErr and eStatCount > 0:
    if vStats:
      print "TOTAL MODEL"
      print "Log likelihood ",
    print edgeLikelihood + nodeLikelihood, 
    if vStats:
      print "\nNull likelihood ",
    print randEdgeLikelihood + randNodeLikelihood,
    if vStats:
      print "\nDeviance",
    print -2.0*(edgeLikelihood+nodeLikelihood),
    if vStats:
      print "\nNull deviance",
    print -2.0*(edgeLikelihood + nodeLikelihood - randEdgeLikelihood - randNodeLikelihood),
    if vStats:
      print "\nPer choice likelihood ratio",
    print math.exp((edgeLikelihood + nodeLikelihood - randEdgeLikelihood - randNodeLikelihood)/(nStatCount+eStatCount))
  elif not vStats:
    print "NA NA NA NA NA"
if outerModFile:
  outerModPtr.close()
nodeFile.close()
edgeFile.close()

