#!/usr/bin/env python

#netmodel class

import sys, random, math

class ModelComponent:
  "This is a component of a network Model"
  MC_TYPE_RANDOM= 1
  MC_TYPE_DEGREE= 2
  MC_TYPE_PFP= 3
  MC_TYPE_SINGLE= 4
  MC_TYPE_DOUBLE= 5
  MC_TYPE_TRI= 6
  MC_TYPE_HOT= 7
  MC_TYPE_RECENT= 8
  MC_TYPE_OLD= 9
   
  def __init__(self):
    self.modelType= 1
    self.modelProb= 0.0
    self.parms=[]
  
  def __init__(self,initParts):
    if len(initParts) < 2:
      print >> sys.stderr,"Not a valid model line",initParts
      sys.exit()
    self.modelType= int(initParts[0])
    self.modelProb= float(initParts[1])
    self.parms=[]
    for f in initParts[2:]:
      self.parms.append(float(f))

class NetModel:
  "A class which represents the statistical model which constructs a network"
  
  NET_MODEL_EPSILON_ = 1e-06
  NET_MODEL_INNER_EDGE = 1
  NET_MODEL_NODE_EDGE = 2
  NET_MODEL_ADD_NODE = 1
  NET_MODEL_ADD_NEW_NODE_EDGE = 2
  NET_MODEL_ADD_INNER_EDGE = 3
    
  def __init__(self):
    self.outerNodeModel= False
    self.outerEdgeModel= False
    self.innerNodeModel= False
    self.innerEdgeModel= False
    self.simpleGraph= False
    self.nodePFPparm= 0
    self.edgePFPparm= 0
    self.nodeHotCount= 0
    self.edgeHotCount= 0

  def read(self,fName):
    try:
      line= "cannot open",fName
      f= open(fName,'r')
      outerEdgeSum= 0.0
      outerNodeSum= 0.0
      innerEdgeSum= 0.0
      innerNodeSum= 0.0
      for line in f.readlines():
#        print 'Line is "',line.strip(),'"'
        if line.strip() == '' or line[0] == '#':
          continue
        parms= line.split()
        if parms[0] == 'n':
          if self.outerNodeModel != False:
            print >> sys.stderr, "Outer node model defined twice"
            sys.exit()
          self.outerNodeModel=[]
          for p in parms[1:]:
            self.outerNodeModel.append(float(p))
            outerNodeSum+= float(p)
          if self.outerNodeModel[0] != 0.0:
            print >> sys.stderr,"Outer node model cannot have ",
            print >> sys.stderr,"zero for first component."
            sys.exit()
        elif parms[0] == 'e':
          if self.outerEdgeModel != False:
            print >> sys.stderr, "Outer edge model defined twice"
            sys.exit()
          self.outerEdgeModel=[]
          for p in parms[1:]:
            self.outerEdgeModel.append(float(p))
            outerEdgeSum+= float(p)
        elif parms[0] == 'N':
          if (self.innerNodeModel == False):
            self.innerNodeModel=[]
          newComp= ModelComponent(parms[1:])
          if newComp.modelType == newComp.MC_TYPE_HOT:
            if self.nodeHotCount != 0:
              print >> sys.stderr, "Can't define two Hot comps in new node model"
              sys.exit()
            if len(newComp.parms) != 1:
              print >> sys.stderr, "Hot component requires parameter"
              sys.exit()
            self.nodeHotCount= int(newComp.parms[0])
          if newComp.modelType == newComp.MC_TYPE_PFP:
            if self.nodePFPparm != 0:
              print >> sys.stderr, "Can't define two PFP comps in new node model"
              sys.exit()
            if len(newComp.parms) != 1:
              print >> sys.stderr, "PFP model requires one parameter"
              sys.exit()
            self.nodePFPparm= newComp.parms[0]
          innerNodeSum+= newComp.modelProb
          self.innerNodeModel.append(newComp)
        elif parms[0] == 'E':
          if (self.innerEdgeModel == False):
            self.innerEdgeModel=[]
          newComp= ModelComponent(parms[1:])
          if newComp.modelType == newComp.MC_TYPE_HOT:
            if self.edgeHotCount != 0:
              print >> sys.stderr, "Can't define two Hot comps in inner edge model"
              sys.exit()
            if len(newComp.parms) != 1:
              print >> sys.stderr, "Hot component requires parameter"
              sys.exit()
            self.edgeHotCount= int(newComp.parms[0])
          if newComp.modelType == newComp.MC_TYPE_PFP:
            if self.edgePFPparm != 0:
              print >> sys.stderr, "Can't define two PFP comps in new node model"
              sys.exit()
            if len(newComp.parms) != 1:
              print >> sys.stderr, "PFP model requires one parameter"
              sys.exit()
            self.edgePFPparm= newComp.parms[0]
          innerEdgeSum+= newComp.modelProb
          self.innerEdgeModel.append(newComp)
        elif parms[0] == 'S' or parms[0] == 's':
          self.simpleGraph= True
        else: 
          raise IOError("Unknown line:",line)
      f.close()
    except Exception, e:
      print >> sys.stderr,e
      print >> sys.stderr,"Problem: ",line
      print >> sys.stderr,fName," is not a valid model file"
      sys.exit()
    if self.innerNodeModel == False or self.outerNodeModel == False:
      print >> sys.stderr, "Must define inner and outer node models"
      sys.exit()
    if abs(innerNodeSum - 1.0) > self.NET_MODEL_EPSILON_:
      print >> sys.stderr, "Inner node model probabilities must total to 1.0"
      print >> sys.stderr,fName," is not a valid model file"
      sys.exit()
    if abs(outerNodeSum - 1.0) > self.NET_MODEL_EPSILON_:
      print >> sys.stderr, "Outer node model probabilities must total to 1.0"
      print >> sys.stderr,fName," is not a valid model file"
      sys.exit()
    if self.outerEdgeModel == False and self.innerEdgeModel == False:
      return
    if self.outerEdgeModel == False or self.innerEdgeModel == False:
      print >> sys.stderr, "Must define either inner and outer edge model or neither"
      sys.exit()
    if abs(innerEdgeSum - 1.0) > self.NET_MODEL_EPSILON_:
      print >> sys.stderr, "Inner edge model probabilities must total to 1.0"
      print >> sys.stderr,fName," is not a valid model file"
      sys.exit()
    if abs(outerEdgeSum - 1.0) > self.NET_MODEL_EPSILON_:
      print >> sys.stderr, "Outer edge model probabilities must total to 1.0"
      print >> sys.stderr,fName," is not a valid model file"
      sys.exit()
  
  def printLine(self, node, choiceNo, picked, choices, setSize, modelSums, model,  network,  fileptr):
    (tDegSum,tSingleSum,tDoubleSum, tTriSum,tPFPSum,tHotSum,tRecSum,tOldSum)= modelSums
    print >> fileptr, node, choiceNo, picked, choices, setSize,
    print >> fileptr, network.degCount[node], network.triCount[node],
    if model == self.NET_MODEL_NODE_EDGE:
      print >> fileptr, network.nodePFP[node],
    else:
      print >> fileptr, network.edgePFP[node],
    print >> fileptr, tDegSum,tSingleSum,tDoubleSum, tTriSum,tPFPSum,
    print >> fileptr, len(network.edges)-network.nodeHeat[node]
  
  def printNodeChoice(self, node, network, choiceNo, fileptr):
    "Print info about node choices from single node to a file"
    model= self.NET_MODEL_NODE_EDGE
    modelSums= self.calcModelSums(network,model,[network.nodeCount-1])
    for i in xrange(0,network.nodeCount-1):
      if i == node:
        pick= 1
      else:
        pick= 0
      self.printLine(i, choiceNo, pick, 1, network.nodeCount-1, modelSums, model, network, fileptr)
  
  def printEdgeChoice(self, node, network, choiceNo, fileptr,skiplist):
    "Print info about node choices from edges to a file"
    model= self.NET_MODEL_INNER_EDGE
    modelSums= self.calcModelSums(network,model,skiplist)
    for i in xrange(0,network.nodeCount):
      if i in skiplist:
        continue
      if i == node:
        pick= 1
      else:
        pick= 0
      self.printLine(i, choiceNo, pick, 1, network.nodeCount - len(skiplist), modelSums, model, network, fileptr)
      
  def printNodeBatch(self, choices, network, choiceNo, fileptr):
    "Print a batch of choices about nodes connecting to new node"
    model = self.NET_MODEL_NODE_EDGE
    choiceSum = sum(choices)
    if choiceSum == 0:
      return
    modelSums= self.calcModelSums(network, model, [network.nodeCount-1])
    for i in xrange(0,network.nodeCount-1):
      self.printLine(i, choiceNo, choices[i], choiceSum, network.nodeCount-1, modelSums, model, network, fileptr)

  def printEdgeBatch(self, choices, network, choiceNo, fileptr):
    "Print a batch of choices about nodes connecting to new node"
    model = self.NET_MODEL_INNER_EDGE
    choiceSum = sum(choices)
    if choiceSum == 0:
      return
    modelSums= self.calcModelSums(network, model, [])
    for i in xrange(0,network.nodeCount):
      self.printLine(i, choiceNo, choices[i], choiceSum, network.nodeCount, modelSums, model, network, fileptr)
  
  def getNprob(self, network, node, model, noNodes, sums):
    "Print the probability of an individual node"
    (tDegSum,tSingleSum,tDoubleSum, tTriSum,tPFPSum,tHotSum,tRecSum,tOldSum)= sums
    if model == self.NET_MODEL_INNER_EDGE:
      thisMod= self.innerEdgeModel
    else:
      thisMod= self.innerNodeModel
    prob= 0.0
    for t in thisMod:
      if t.modelType == t.MC_TYPE_RANDOM:
        cont= 1.0/noNodes
      elif t.modelType == t.MC_TYPE_DEGREE:
        if tDegSum == 0:
          cont= 1.0/noNodes
        else:
          cont= float(network.degCount[node])/tDegSum
      elif t.modelType == t.MC_TYPE_PFP:
        if tPFPSum == 0.0:
          cont= 1.0/noNodes
        elif model == self.NET_MODEL_INNER_EDGE:
          cont= network.edgePFP[node]/tPFPSum
        else:
          cont= network.nodePFP[node]/tPFPSum
      elif t.modelType == t.MC_TYPE_SINGLE:
        if tSingleSum == 0:
          cont= 1.0/noNodes
        elif network.degCount[node] == 1:
          cont= 1.0/tSingleSum
        else:
          cont= 0.0
      elif t.modelType == t.MC_TYPE_DOUBLE:
        if tDoubleSum == 0:
          cont= 1.0/noNodes
        elif network.degCount[node] == 2:
          cont= 1.0/tDoubleSum
        else:
          cont= 0.0
      elif t.modelType == t.MC_TYPE_TRI:
        if tTriSum == 0:
          cont= 1.0/noNodes
        else:
          cont= float(network.triCount[node])/tTriSum
      elif t.modelType == t.MC_TYPE_HOT:
        if tHotSum == 0:
          cont= 1.0/noNodes
        elif (len(network.edges) - network.nodeHeat[node]) < t.parms[0]:
          cont= 1.0/tHotSum
        else:
          cont= 0.0
      elif t.modelType == t.MC_TYPE_RECENT:
        if tRecSum == 0:
          cont= 1.0/noNodes
        elif node >= network.nodeCount-t.parms[0]:
          cont= 1.0/tRecSum
        else:
          cont= 0.0
      elif t.modelType == t.MC_TYPE_OLD:
        cont= float(network.nodeCount-node)/tOldSum
      else:
        print >> sys.stderr, "Unknown model type in getNprob"
        sys.exit()
      prob+= t.modelProb*cont
      
    return prob
    
  def getNodeProb(self, node, network):
    "Calculate probabiltiy of node connecting to new node"
    skiplist= [network.nodeCount-1]
    model= self.NET_MODEL_NODE_EDGE
    if self.simpleGraph:
      skiplist.extend(network.outNodes[network.nodeCount-1])
    sums= self.calcModelSums(network, model ,skiplist)  
    nNodes= network.nodeCount-len(skiplist)
    prob= self.getNprob(network, node, model, nNodes ,sums)
    rprob= 1.0/nNodes
    return (prob,rprob)
    
  def getEdgeProb(self, node1, node2, network):
    "Calculate probabiltiy of node connecting to inner edge nodes"
    model= self.NET_MODEL_INNER_EDGE
    sums= self.calcModelSums(network, model ,[]) 
    nNodes= network.nodeCount
    p1= self.getNprob(network, node1, model, nNodes ,sums)
    rp1= 1.0/nNodes
    skiplist=[node1]
    if self.simpleGraph:
      skiplist.extend(network.outNodes[node1])

    nNodes= network.nodeCount-len(skiplist)
    sums= self.calcModelSums(network, model ,skiplist) 
    p2= self.getNprob(network, node2, model, nNodes ,sums)
    rp2= 1.0/nNodes
    prob= p1*p2
    rprob= rp1*rp2
    return (prob,rprob)
    
  def calcModelSums(self, network, model, skiplist):
    "Calculate sums over various network parameters"
    tDegSum= network.degSum
    tSingleSum= network.singleSum
    tDoubleSum= network.doubleSum
    tTriSum= network.triSum
    tPFPSum= 0.0
    tHotSum= 0
    tRecSum= 0
    tOldSum= 0
    PFP= False
    hotCount= 0
    recCount= 0
    if model == self.NET_MODEL_NODE_EDGE:
      if self.nodePFPparm != 0:
        PFP= True
        tPFPSum= sum(network.nodePFP)
      if self.nodeHotCount != 0:
        hotCount= self.nodeHotCount
    elif model == self.NET_MODEL_INNER_EDGE:
      if self.edgePFPparm != 0:
        PFP= True
        tPFPSum= sum(network.edgePFP)
      if self.edgeHotCount != 0:
        hotCount= self.edgeHotCount
    tOldSum= (network.nodeCount-1)*(network.nodeCount-2)/2
    nEdges= len(network.edges)
    if hotCount != 0:
      for i in range(max(0,nEdges-hotCount), nEdges):
        e= network.edges[i]
        if network.nodeHeat[e[0]] == i+1 and e[0] not in skiplist:
          tHotSum+= 1
#          print >> sys.stderr,e[0],
        if network.nodeHeat[e[1]] == i+1 and e[0] != e[1] and e[1] not in skiplist:
          tHotSum+= 1
#          print >> sys.stderr,e[1],
#      print >> sys.stderr,"thotsum",tHotSum
    if not self.simpleGraph:
      skipList= set(skiplist)
    for s in skiplist:
      tDegSum-= network.degCount[s]
      if network.degCount[s] == 1:
        tSingleSum-= 1
      elif network.degCount[s] == 2:
        tDoubleSum-= 1
      tTriSum-= network.triCount[s]
      tOldSum-= s
#        print >> sys.stderr,"Remove",s
      if PFP:
        if model == self.NET_MODEL_INNER_EDGE:
          tPFPSum-= network.edgePFP[s]
        else:
          tPFPSum-= network.nodePFP[s]
    if tHotSum < 0: 
      print >> sys.stderr, "Error in thotsum skiplist=",skiplist
      sys.exit()
#    print >> sys.stderr,"tHotSum",tHotSum
    return (tDegSum,tSingleSum,tDoubleSum, tTriSum,tPFPSum,tHotSum,tRecSum,
        tOldSum)
          
  def pickNode(self, network, model, skiplist):
    tNodes= network.nodeCount - len(skiplist)
    # Check if any nodes available
    if tNodes <= 0:
      return -1
    sums= self.calcModelSums(network, model, skiplist)
    nNodes= network.nodeCount-len(skiplist)
    dSum= 0.0
    rNo= random.random()
    for i in xrange(0,network.nodeCount):
      if i in skiplist:
        continue
      dSum+= self.getNprob(network,i,model,tNodes,sums)
      if rNo < dSum:
        return i
    print >> sys.stderr,"oops"
    for i in xrange(network.nodeCount-1,-1,-1):
      if not i in skiplist:
        return i
    print >> sys.stderr,"Problem in pickNode routine"
  
  def pickNodeEdge(self, network):
    skiplist= [network.nodeCount-1]
    if self.simpleGraph == True:
      skiplist.extend(network.outNodes[network.nodeCount-1])
    node= self.pickNode(network, self.NET_MODEL_NODE_EDGE, skiplist)
    return node
  
  def pickInnerEdge(self, network):
    node1= self.pickNode(network, self.NET_MODEL_INNER_EDGE, [])
    if node1 == -1:
      return -1
    skiplist= [node1]
    if self.simpleGraph == True:
      skiplist.extend(network.outNodes[node1])
#    print >> sys.stderr,"skiplist ==",skiplist
    node2= self.pickNode(network, self.NET_MODEL_INNER_EDGE, skiplist)
    if node2 == -1:
      return -1
    return (node1, node2)
  
  def pickFromDist(self,dist):
    "Dist is list representing a distribution"
    count= 0
    distSum= 0.0
    r= random.random()
    j= 0
    for i in dist:
      distSum+= i
      if r < distSum:
        return j
      j+= 1
    return j-1
  
  def performIterOuterFile(self, network, outerModPtr):
    "iterate network model with outer model from file"
    
    count= 0
    for i in outerModPtr.readlines():
      try:
        op= int(i.split()[0])
      except:
        print >> sys.stderr,"Badly formatted line in outer mod file",i
        sys.exit()
      if op == self.NET_MODEL_ADD_NODE:
        network.addNode(network.nodeCount)
      elif op == self.NET_MODEL_ADD_NEW_NODE_EDGE:
        newEdgeNode= self.pickNodeEdge(network)
        if newEdgeNode != -1:
          network.addEdge(network.nodeCount-1,newEdgeNode,count,self.nodePFPparm, self.edgePFPparm)
      elif op == self.NET_MODEL_ADD_INNER_EDGE:
        newEdge= self.pickInnerEdge(network)
        if newEdge != -1:
          network.addEdge(newEdge[0],newEdge[1],count,self.nodePFPparm,self.edgePFPparm)
      count+= 1
  
  def performIter(self,network,noIter):
    "iterate the network model for a certain number of times"
    nodeCount= self.pickFromDist(self.outerNodeModel)
    edgeCount= 0
    newNode= True
#    print network.nodes
#    print network.nodeCount
#    for i in range (0, network.nodeCount):
#      print network.outNodes[i]
    for i in xrange(0, noIter):
      if newNode == True:
        network.addNode(network.nodeCount)
        newNode= False
      if nodeCount > 0:
        newEdgeNode= self.pickNodeEdge(network)
        if newEdgeNode != -1:
          network.addEdge(network.nodeCount-1,newEdgeNode,i,self.nodePFPparm,self.edgePFPparm)
        nodeCount-= 1
        if nodeCount == 0:
          if self.outerEdgeModel != False: 
            edgeCount= self.pickFromDist(self.outerEdgeModel)
            if edgeCount == 0:
              nodeCount= self.pickFromDist(self.outerNodeModel)
              newNode= True
          else:
            nodeCount= self.pickFromDist(self.outerNodeModel)
            newNode= True
      elif edgeCount > 0:
        newEdge= self.pickInnerEdge(network)
        if newEdge != -1:
          network.addEdge(newEdge[0],newEdge[1],i,self.nodePFPparm,self.edgePFPparm)
        edgeCount-= 1
        if edgeCount == 0:
          nodeCount= self.pickFromDist(self.outerNodeModel)
          newNode= True
      else:
        print >> sys.stderr,"Shouldn't get here!"
        sys.exit()
    
class NetData:
  "A class representing a network"
  
  def __init__(self,model= False):
# Initialise with a base network
    self.initCounters()

    self.edges=[]
    self.nodeCount= 0
    self.nodes={}
    self.addNode(0)
    self.addNode(1)
    if model == False:
      self.addEdge(1,0,1,0,0)
    else:
      self.addEdge(1,0,1,model.nodePFPparm,model.edgePFPparm)
  
  def initCounters(self):
    self.degCount= []
    self.degSum= 0
    self.nodePFP= []
    self.edgePFP= []
    self.singleSum= 0
    self.doubleSum= 0
    self.triCount= []
    self.triSum= 0
    self.outNodes=[]
    self.nodeHeat=[]
    

  def addTempEdge(self,e1,e2,time,overRide,model):
    "Adds edge to network if it will leave network connected"
    "Over ride adds edge anyway even if both nodes are new"
    "Returns True if edge was added" 
#    print self.nodes
#    print "Trying",(e1,e2)
    try:
      n1= self.nodes[e1]
    except:
      n1= -1
    try:
      n2= self.nodes[e2]
    except: 
      n2= -1
#   This occurs if both nodes are new and 
    if n1 == -1 and n2 == -1 and overRide == False:
      return False
    if n2 == -1:
      if model == False:
        self.addSimpleNode(e2)
      else:
        self.addNode(e2)
      n2= self.nodeCount-1
    if n1 == -1:
      if model == False:
        self.addSimpleNode(e1)
      else:
        self.addNode(e1)
      n1= self.nodeCount-1
    if model == False:
      self.addSimpleEdge(max(n1,n2),min(n1,n2),time)
    else: 
      self.addEdge(max(n1,n2),min(n1,n2),time, model.nodePFPparm, model.edgePFPparm)
    return True
  
  def addSimpleNode(self, nodeNo):
    self.nodes[nodeNo]= self.nodeCount
    self.nodeCount+= 1
    
  def addNode(self, nodeNo):
    "Add a new node into the network"
    self.nodes[nodeNo]= self.nodeCount
    self.nodeCount+= 1
    self.degCount.append(0)
    self.nodePFP.append(0)
    self.edgePFP.append(0)
    self.triCount.append(0)
    self.nodeHeat.append(0)
    self.outNodes.append(list())
  
  def addSimpleEdge(self, e1, e2, time):
    "Add a new edge into the network without updating counts"
    self.edges.append((e1,e2,time))
    
  def addEdge(self, e1, e2, time, nodePFPparm, edgePFPparm):
    "Add a new edge into the network"
    self.degCount[e1]+= 1
    self.degCount[e2]+= 1
    self.degSum+= 2
    if (self.degCount[e1] == 1):
      self.singleSum+= 1
    if (self.degCount[e2] == 1):
      self.singleSum+= 1
    if (self.degCount[e1] == 2):
      self.singleSum-= 1
      self.doubleSum+= 1
    if (self.degCount[e2] == 2):
      self.singleSum-= 1
      self.doubleSum+= 1
    if (self.degCount[e1] == 3):
      self.doubleSum-= 1
    if (self.degCount[e2] == 3):
       self.doubleSum-= 1
    for i in self.outNodes[e1]:
      if i in self.outNodes[e2]:
        self.triCount[e1]+= 1
        self.triCount[e2]+= 1
        self.triCount[i]+= 1
        self.triSum+= 3
        break
    self.outNodes[e1].append(e2)
    self.outNodes[e2].append(e1)
    if nodePFPparm != 0:
      d= self.degCount[e1]
      self.nodePFP[e1]= pow(d,1.0+nodePFPparm * math.log10(d))
      d= self.degCount[e2]
      self.nodePFP[e2]= pow(d,1.0+nodePFPparm * math.log10(d))
    if edgePFPparm != 0:
      d= self.degCount[e1]
      self.edgePFP[e1]= pow(d,1.0+edgePFPparm * math.log10(d))
      d= self.degCount[e2]
      self.edgePFP[e2]= pow(d,1.0+edgePFPparm * math.log10(d))        
    self.edges.append((e1,e2,time))
    if e1 != self.nodeCount-1:
      self.nodeHeat[e1]= len(self.edges)
    if e2 != self.nodeCount-1:
      self.nodeHeat[e2]= len(self.edges)
  
  def read(self,netFileName, model= False, dropDis= False, ignoreDis=False, forceSimple=False):
    "Read network from a file"
    "Drop dis simply removes all non-connected links"
    "ignoreDis ignores any disconnected at the end"
    self.edges=[]
    self.nodeCount= 0
    self.nodes= {}   # temporary node dictionary
    self.initCounters()
    time= 0
    firstNode= True
    if forceSimple:
      edgeList={}
    try:
      netFileDone= False
      netFile= open(netFileName,'r')
      waitList= []
      while (1):
        if netFileDone and len(waitList) == 0:
          return
        waitFound= False
        for i in xrange(0, len(waitList)):
          if self.addTempEdge(waitList[i][0],waitList[i][1],time,False, model):
#            print >> sys.stderr, "Edge",waitList[i][0],",",waitList[i][1]," moved",
#            print >> sys.stderr, "in time from",waitList[i][2],"to",time
#            print >> sys.stderr,"Popping",i,len(waitList)
            waitList.pop(i)
#            print >> sys.stderr,"Popped",len(waitList)
            waitFound= True
            break
        if waitFound:
          continue
#        if len(waitList) != 0:
#          print waitList
        if netFileDone:
          
          if not ignoreDis:
            print >> sys.stderr, waitList
            print >> sys.stderr, len(waitList),"edges remain disconnected"
            sys.exit()
          print >> sys.stderr, len(waitList),"edges dropped from analysis"
          return
        line= netFile.readline()
#        print >> sys.stderr,"Read", line,
        if line == '':
          netFile.close()
          netFileDone= True
          continue
        if line[0] == '#':
          continue
        if line.strip() == "":
          continue
        split= line.split()
        if len(split) < 2:
          print >> sys.stderr, "Invalid line"+line
          raise Exception("Invalid line"+line)
        e1= int(split[0])
        e2= int(split[1])
        if len(split) <= 2:
          time+= 1
        else:
          time= int(split[2])
        if forceSimple:
          if e1 == e2:
            continue
          e= (max(e1,e2),min(e1,e2))
          if edgeList.has_key(e):
            continue
          edgeList[e]= True
        if firstNode:
          edgeAdded= self.addTempEdge(e1,e2,time,True,model)
          firstNode= False
        else:
          edgeAdded= self.addTempEdge(e1,e2,time,False,model)
        
        if not edgeAdded and not dropDis:
          waitList.append(((e1,e2),time))
      
    except Exception, e:
      print >> sys.stderr,"Error:",e
      print >> sys.stderr,netFileName,"appears not to be a valid network"
      sys.exit()

  def printOut(self):
    "print the edges on the network"
    for e in self.edges:
      if e[0] > e[1]:
        print e[0],e[1]
      else:
        print e[1],e[0]

